/*
 * traceroute implementation that does not require root privilege.
 *
 * This requires a Linux 2.4 kernel.
 *
 * Copyright (C) 2000-2008, Olaf Kirch <okir@suse.de>
 *
 * License: GPL 2.0 or later. See COPYING.GPL for details.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <netinet/icmp6.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <locale.h>
#include <assert.h>
#include <fcntl.h>
#include "utils.h"
#include "version.h"

/* The kernel wants these */
typedef u_int32_t	__u32;
typedef u_int8_t	__u8;
#include <linux/errqueue.h>

/* These are not available with glibc 2.1.3: */
#ifndef SO_TIMESTAMP
#define SO_TIMESTAMP	29
#endif
#ifndef IPV6_RECVERR
#define IPV6_RECVERR	25
#endif

/*
 * older glibc compatibility
 * From rfc 3542:
 * -  Changed the name of ICMPv6 unreachable code 2 to be "beyond scope
 *    of source address."  ICMP6_DST_UNREACH_NOTNEIGHBOR was removed
 *    with this change.
 */
#if defined(ICMP6_DST_UNREACH_NOTNEIGHBOR) && !defined(ICMP6_DST_UNREACH_BEYONDSCOPE)
# define ICMP6_DST_UNREACH_BEYONDSCOPE ICMP6_DST_UNREACH_NOTNEIGHBOR
#endif

#define MAXHOPS		255
#define MAXPROBES	11
#define MAXGATEWAYS	9

struct hop {
	int		fd;
	unsigned int	ttl;
	unsigned int	sent;
	unsigned int	recvd;
	unsigned int	final;

	/* Previously printed address */
	struct sockaddr_storage	prev_addr;

	struct probe {
	    struct hop *	hop;
	    int			fd;
	    struct timeval	sent_time;
	    unsigned long	timeout;
	    struct timeval	recvd_time;
	    struct sockaddr_storage responder;
	    const char *	err_ind;
	    unsigned int	done : 1,
	    			printed : 1;

	    /* This is for ICMP */
	    uint16_t		seq;
	    uint32_t		hdr[64];
	    unsigned int	hdrlen;
	}		probe[MAXPROBES];
};

struct response {
	void *			buf;
	size_t			bufsz;
	struct sock_extended_err ee;
	struct sockaddr_storage	sender;
	struct timeval		ts;
};

struct protocol {
	int		ipproto;
	const char *	name;
	uint16_t	dst_port;
	int		poll_mask;

	void		(*init)(void);
	int		(*init_probe)(struct hop *, struct probe *);
	int		(*send_probe)(struct probe *);
	void		(*recv_probe)(struct probe *);
	struct probe *	(*find_probe)(void *pkt, unsigned int pktlen,
					struct probe **list, unsigned int len);
};

static struct hop	hops[MAXHOPS];
static unsigned int	first_hop, last_hop;
static unsigned int	opt_timeout = 5,
			opt_retries = 3,
			concurrent_hops = 6,
			max_hops = 30;
static unsigned int	have_final_response;
static unsigned long	now;

static int		af = -1;
static size_t		af_len = 0;
static char *		src_addr_name;
static struct sockaddr_storage src_addr;
static const char *	src_device;
static u_int16_t	src_port = 64000;
static const char *	dst_name;
static struct sockaddr_storage dst_addr;
static u_int16_t	dst_port = 0;
static u_int8_t		dst_tos;

static int		opt_debug = 0;
static int		opt_noresolver = 0;
static int		opt_dontfrag = IP_PMTUDISC_DONT;
static int		opt_dontroute = 0;
static int		opt_rr = 0;
static char *		opt_gateway_name[MAXGATEWAYS];
static struct sockaddr_storage opt_gateway[MAXGATEWAYS];
static unsigned int	ngateways;

static unsigned char	packet[65536];
static unsigned int	packetsize = 40;
static unsigned char	ipoptions[40];
static unsigned int	ipoptions_len = 0;

static int		icmp_sendsock = -1;
static int		icmp_recvsock = -1;
static uint16_t		icmp_ident = 0;
static uint16_t		icmp_sequence = 0;

static void		set_now(void);
static void		hop_next(void);
static void		hop_xmit(struct hop *);
static void		hop_print(struct hop *);
static void		probe_reduce_timeout(unsigned int ttl);

static void		icmp_init_recvsock(void);
static int		icmp_recv(struct probe **, int);

static void		ipopt_init(void);
static void		usage(int exval);
static void		show_version(void);

static struct protocol	trace_udp;
static struct protocol	trace_icmp;
static struct protocol	trace_tcp;

static struct protocol *proto = &trace_udp;

#define dprint(args...)	do { \
		if (opt_debug >= 1) printf(args); \
	} while (0)
#define dprint2(args...)	do { \
		if (opt_debug >= 2) printf(args); \
	} while (0)
#define dprint3(args...)	do { \
		if (opt_debug >= 3) printf(args); \
	} while (0)

int
main(int argc, char **argv)
{
	unsigned int	n;
	int		c, warn_i_change = 0;

	/* Set the locale - needed for international domain names */
	setlocale (LC_ALL, "");

	while ((c = getopt(argc, argv, "46DFVf:g:i:Im:N:np:t:w:q:rRS:T")) != -1) {
		if ((c == 'i' || c == 'I') && !warn_i_change) {
			fprintf(stderr,
				"Note: the -i and -I options were exchanged"
				"for compability with LBL traceroute\n"
				"Use -I for ICMP, and -i <ifname> to specify the interface name\n");
			warn_i_change = 1;
		}

		switch (c) {
		case '4':
			af = AF_INET;
			break;
		case '6':
			af = AF_INET6;
			break;
		case 'D':
			opt_debug++;
			break;

		case 'F':
			opt_dontfrag = IP_PMTUDISC_DO;
			break;

		case 'f':
			first_hop = getnum("hop", optarg, 1, 255);
			break;

		case 'g':
			/* fatal("option -g not yet supported"); */
			if (ngateways >= MAXGATEWAYS)
				fatal("too many gateways");
			opt_gateway_name[ngateways++] = optarg;
			break;

		case 'I':
			proto = &trace_icmp;
			break;

		case 'i':
			src_device = optarg;
			break;

		case 'm':
			max_hops = getnum("hop", optarg, 1, 255);
			break;

		case 'N':
			concurrent_hops = getnum("concurrent hop", optarg, 1, 16);
			break;

		case 'n':
			opt_noresolver = 1;
			break;

		case 'p':
			dst_port = getnum("port", optarg, 1, 65535);
			break;

		case 't':
			dst_tos = getnum("TOS", optarg, 0, 255);
			break;

		case 'w':
			opt_timeout = getnum("timeout", optarg, 1, 256);
			break;

		case 'q':
			opt_retries = getnum("retry", optarg, 1, MAXPROBES - 1);
			break;

		case 'r':
			opt_dontroute = 1;
			break;

		case 'R':
			opt_rr = 1;
			break;

		case 'S':
			src_addr_name = optarg;
			break;

		case 'T':
			proto = &trace_tcp;
			break;

		case 'V':
			show_version();
			break;
		default:
			usage(1);
		}
	}

	if (optind == argc - 2) {
		packetsize = getnum("packet length", argv[optind+1],
					1, 65536);
	} else
	if (optind != argc - 1)
		usage(1);

	/* Check the program name. If we've called as traceroute6,
	 * default to IPv6 */
	if (af == -1) {
		char	*s;

		if ((s = strrchr(argv[0], '/')) != 0)
			s++;
		else
			s = argv[0];
		if (!strcmp(s, "traceroute6"))
			af = AF_INET6;
	}
	if (af == AF_INET)
		af_len = sizeof(struct sockaddr_in);
	else
		af_len = sizeof(struct sockaddr_in6);

	dst_name = argv[optind];
	if (!getaddr(dst_name, af, &dst_addr))
		return 1;
	af = dst_addr.ss_family;

	if (!dst_port)
		dst_port = proto->dst_port;

	/* Resolve gateways if given */
	for (n = 0; n < ngateways; n++) {
		if (!getaddr(opt_gateway_name[n], af, opt_gateway + n))
			return 1;
	}

	/* Resolve source address if given */
	if (src_addr_name && !getaddr(src_addr_name, af, &src_addr)) {
		return 1;
	}

	if (first_hop >= max_hops) {
		fprintf(stderr, "first hop %u larger than max hops %u\n",
			first_hop, max_hops);
		return 1;
	}

	/* Initialize packet */
	for (n = 0; n < packetsize; n++)
		packet[n] = 0x40 + (n & 0x3f);

	/* Initialize IP options */
	ipopt_init();

	/* Initialize the protocol hooks */
	if (proto->init)
		proto->init();

	/* Start by sending the first packet */
	last_hop = first_hop;

	printf("traceroute to %s (%s), %u hops max, %u byte packets using %s\n",
		dst_name, straddr(&dst_addr),
		max_hops, packetsize,
		proto->name);

	while (1) {
		struct probe	*probe[MAXHOPS*MAXPROBES], *pb;
		struct pollfd	pfd[MAXHOPS*MAXPROBES];
		unsigned long	timeout;
		struct hop	*hop;
		unsigned int	nfds, nprobes, m, n, wait_some = 0;

		while (last_hop < max_hops
		    && last_hop < first_hop + concurrent_hops)
			hop_next();

		set_now();
		timeout = now + opt_timeout * 1000;
		memset(pfd, 0, sizeof(pfd));

		nfds = nprobes = 0;
		for (hop = hops + first_hop; hop < hops + last_hop; hop++) {
			unsigned int	num;

			for (num = hop->recvd; num < opt_retries; num++) {
				pb = hop->probe + num;

				if (num >= hop->sent) {
					hop_xmit(hop);
					wait_some = 1;
				}

				if (pb->fd >= 0 && proto->poll_mask) {
					pfd[nfds].events = proto->poll_mask;
					pfd[nfds].fd = pb->fd;
					nfds++;
				}
				if (pb->timeout < timeout)
					timeout = pb->timeout;
				probe[nprobes++] = pb;

				/* If we've already transmitted some
				 * packets, wait for 1ms to receive
				 * something. This helps us give much more
				 * accurate RTTs on fast links */
				if (wait_some) {
					timeout = now + 1;
					goto wait_for_some;
				}
			}
		}

		if (nprobes == 0)
			break;

wait_for_some:
		if (icmp_recvsock >= 0) {
			pfd[nfds].events = POLLERR;
			pfd[nfds].fd = icmp_recvsock;
			nfds++;
		}

		dprint2("poll %d fds, timeout %lu ms\n", nfds, timeout - now);
		poll(pfd, nfds, (now < timeout)? (timeout - now) : 0);

		set_now(); /* time has passed - catch up */

		/* Receive any pending ICMP errors */
		for (m = n = 0; n < nprobes; n++) {
			pb = probe[n];
			if (pb->fd >= 0 && pfd[m++].revents) {
				proto->recv_probe(pb);

				/* When we've received a respone from a hop,
				 * reduce the timeout for all hops that have
				 * the same distance or less.
				 */
				if (pb->done)
					probe_reduce_timeout(pb->hop->ttl);
			} else if (pb->timeout <= now) {
				/* timeout - mark as done. */
				pb->done = 1;
			}
		}
		if (icmp_recvsock >= 0) {
			while (icmp_recv(probe, nprobes) >= 0)
				;
		}

		/* FIXME: we really shouldn't start printing packets
		 * and resoving addresses while we still have packets
		 * in flight, especially on low RTT links.
		 * We should either skip the output part while that's
		 * the case, or fork and do the printing in a separate
		 * process */

		/* Now loop over all hop structures and see whether we can
		 * add more probes, or wrap up the timed out ones */
		while (first_hop < last_hop) {
			hop = hops + first_hop;
			hop_print(hop);
			if (hop->recvd < hop->sent || hop->sent < opt_retries)
				break;

			/* This one is complete */
			printf("\n");
			first_hop++;
			if (hop->final)
				goto done;
		}
	}

done:	return 0;
}

static void
usage(int exval)
{
	fprintf(stderr,
	"usage: traceroute [-nFV] [-f first_ttl] [-m max_hops] [-p port]\n"
	"           [-S source_addr] [-I interface] [-g gateway]\n"
	"           [-t tos] [-w timeout] [-q nqueries] host [packetlen]\n");
	exit(1);
}

static void
show_version(void)
{
	printf("This is traceroute %s\n", TRACEROUTE_VERSION);
	printf("Copyright (c) 2000-2008 Olaf Kirch\n");
	exit(0);
}

/* Set our idea of the current time */
static void
set_now(void)
{
	struct timeval tv;

	gettimeofday(&tv, NULL);
	now = tv.tv_sec * 1000 + tv.tv_usec / 1000;
}

static void
hop_next(void)
{
	struct hop	*hop = hops + last_hop++;

	hop->ttl = last_hop;
}

static void
hop_xmit(struct hop *hop)
{
	struct probe	*pb;

	dprint("Sending hop %ld packet %d ttl=%d\n",
			(long)(hop - hops + 1), hop->sent, hop->ttl);

	/* Init probe and send packet */
	pb = hop->probe + hop->sent++;
	proto->init_probe(hop, pb);
	if (proto->send_probe(pb) < 0)
		perror("send failed");
	gettimeofday(&pb->sent_time, NULL);
}

static void
hop_print(struct hop *hop)
{
	struct probe	*p = hop->probe;
	struct timeval	delta;
	unsigned int	nr;

	while (hop->recvd < hop->sent) {
		if (!hop->probe[hop->recvd].done)
			break;
		hop->recvd++;
	}

	for (nr = 0; nr < hop->sent; nr++, p++) {
		if (!p->done)
			break;
		if (p->printed)
			continue;
		if (nr == 0)
			printf("%2lu ", (hop - hops) + 1UL);
		if (p->responder.ss_family == 0) {
			printf(" *");
		} else {
			printf(" ");
			timersub(&p->recvd_time, &p->sent_time, &delta);
			if (delta.tv_sec < 0)
				timerclear(&delta);
			if (!sameaddr(&p->responder, &hop->prev_addr))
				printaddr(&p->responder, !opt_noresolver);
			hop->prev_addr = p->responder;
			if (p->err_ind)
				printf("(%s)", p->err_ind);
			printf("  %lu.%03lu ms",
				1000 * delta.tv_sec + delta.tv_usec / 1000,
				delta.tv_usec % 1000);
		}
		p->printed = 1;
	}
	fflush(stdout);
}

static void
probe_seterror(struct probe *p, struct sock_extended_err *ep)
{
	static char	errbuf[32];
	struct hop	*hop = p->hop;
	const char	*errstring = NULL;

	/* Handle error codes */
	errstring = errbuf;
	if (ep->ee_origin == 0) {
		errstring = NULL;
	} else
	if (ep->ee_origin == SO_EE_ORIGIN_ICMP) {
		switch (ep->ee_type) {
		case ICMP_TIME_EXCEEDED:
			return;

		case ICMP_DEST_UNREACH:
			switch (ep->ee_code) {
			case ICMP_UNREACH_NET:
			case ICMP_UNREACH_NET_UNKNOWN:
			case ICMP_UNREACH_ISOLATED:
			case ICMP_UNREACH_NET_PROHIB:
			case ICMP_UNREACH_TOSNET:
			case ICMP_UNREACH_FILTER_PROHIB:
				errstring = "N!";
				break;

			case ICMP_UNREACH_HOST:
			case ICMP_UNREACH_HOST_UNKNOWN:
			case ICMP_UNREACH_HOST_PROHIB:
			case ICMP_UNREACH_TOSHOST:
			case ICMP_UNREACH_HOST_PRECEDENCE:
				errstring = "H!";
				break;

			case ICMP_UNREACH_PORT:
				/* we've reached the destintation host */
				errstring = 0;
				break;

			case ICMP_UNREACH_PROTOCOL:
				errstring = "P!";
				break;

			case ICMP_UNREACH_NEEDFRAG:
				errstring = "F!";
				break;

			case ICMP_UNREACH_SRCFAIL:
				errstring = "S!";
				break;

			default:
				errstring = "!!";
				break;
			}
			break;
		}
	} else
	if (ep->ee_origin == SO_EE_ORIGIN_ICMP6) {
		switch (ep->ee_type) {
		case ICMP6_TIME_EXCEEDED:
			if (ep->ee_code == ICMP6_TIME_EXCEED_TRANSIT)
				return;
			break;

		case ICMP6_DST_UNREACH:
			switch (ep->ee_code) {
			case ICMP6_DST_UNREACH_NOROUTE:
			case ICMP6_DST_UNREACH_ADMIN:
				errstring = "N!";
				break;

			case ICMP6_DST_UNREACH_BEYONDSCOPE:
			case ICMP6_DST_UNREACH_ADDR:
				errstring = "H!";
				break;

			case ICMP6_DST_UNREACH_NOPORT:
				/* we've reached the destintation host */
				errstring = 0;
				break;

			default:
				errstring = "!!";
				break;
			}
			break;
		}
	}

	if (errstring == errbuf) {
		snprintf(errbuf, sizeof(errbuf), "icmp-%d-%d",
				ep->ee_type, ep->ee_code);
		errstring = strdup(errbuf);
	}

	p->err_ind = errstring;
	have_final_response = 1;
	hop->final = 1;
}

static int
probe_recverr(int fd, struct response *r, int flags)
{
	struct msghdr	msg;
	unsigned char	control[1024];
	struct iovec	iov;
	struct cmsghdr	*cm;
	int		len;

	memset(&msg, 0, sizeof(msg));
	gettimeofday(&r->ts, NULL);

	msg.msg_name	= &r->sender;
	msg.msg_namelen	= sizeof(r->sender);
	msg.msg_control	= control;
	msg.msg_controllen = sizeof(control);
	if (r->buf) {
		iov.iov_base = r->buf;
		iov.iov_len  = r->bufsz;
		msg.msg_iov  = &iov;
		msg.msg_iovlen = 1;
	}

	len = recvmsg(fd, &msg, flags);
	if (len < 0)
		return -1;

	if (flags == 0)
		return len;

	dprint3("cmsgs:");
	for (cm = CMSG_FIRSTHDR(&msg);
	     cm;
	     cm = CMSG_NXTHDR(&msg, cm)) {
		if (cm->cmsg_level == SOL_SOCKET
		 && cm->cmsg_type == SO_TIMESTAMP) {
			dprint3(" SO_TIMESTAMP");
			memcpy(&r->ts, CMSG_DATA(cm), sizeof(r->ts));
			continue;
		}

		if ((cm->cmsg_level == SOL_IP && cm->cmsg_type == IP_RECVERR)
		 || (cm->cmsg_level == SOL_IPV6 && cm->cmsg_type == IPV6_RECVERR)) {
			struct sock_extended_err *ep;

			dprint3(" IP_RECVERR");

			/* the offender follows the extender_err struct;
			 * it's not part of the struct itself. So we copy
			 * this separately. */
			ep = (struct sock_extended_err *) CMSG_DATA(cm);
			memcpy(&r->sender,
				SO_EE_OFFENDER(ep),
				sizeof(struct sockaddr_storage));
			r->ee = *ep;
			continue;
		}

		dprint3(" lvl=%d/typ=%d", cm->cmsg_level, cm->cmsg_type);
	}
	dprint3("\n");

	return len;
}

static void
common_recv_probe(struct probe *p)
{
	struct response	response;
	struct hop	*hop = p->hop;
	int		len;

	/* recv errmsg */
	memset(&response, 0, sizeof(response));
	len = probe_recverr(p->fd, &response, MSG_ERRQUEUE);
	if (len < 0)
		return;

	p->recvd_time = response.ts;
	p->responder = response.sender;
	close(p->fd);
	p->fd = -1;
	p->done = 1;

	dprint("Received hop %u response\n", hop->ttl);

	probe_seterror(p, &response.ee);
}

/*
 * When we receive a response from a router or host N hops
 * away, we assume that other boxes on the path will have
 * a similar delay. Therefore, we reduce the timeout of all
 * pending probes with less or equal TTL to 1 second.
 */
static void
probe_reduce_timeout(unsigned int ttl)
{
	unsigned long timeo = now + 1000;
	struct hop *hop;
	struct probe *pb;
	unsigned int i;

	for (hop = hops + first_hop; hop < hops + last_hop; hop++) {
		if (hop->ttl > ttl)
			break;

		for (pb = hop->probe, i = 0; i < opt_retries; ++i, ++pb) {
			if (!pb->done && timeo < pb->timeout)
				pb->timeout = timeo;
		}
	}
}

/*
 * Generic IPv4/IPv6 address accessor functions
 */
static void af_fatal()
{
	fatal("Unsupported address family %u\n", af);
}

static socklen_t af_addrlen(void)
{
	switch (af) {
	case AF_INET:
		return sizeof(struct sockaddr_in);
	case AF_INET6:
		return sizeof(struct sockaddr_in6);
	}
	af_fatal();
	return 0;
}

static void af_setport(struct sockaddr *sap, uint16_t port)
{
	switch (af) {
	case AF_INET:
		((struct sockaddr_in *) sap)->sin_port = htons(port);
		return;
	case AF_INET6:
		((struct sockaddr_in6 *) sap)->sin6_port = htons(port);
		return;
	}
	af_fatal();
}

/*
 * Helper functions for dissecting an ICMP packet.
 */
static void *
ip_get_header(int protocol, void *pkt, unsigned int len, unsigned int *off)
{
	unsigned char	*raw = (pkt + *off);

	if (protocol == IPPROTO_IP) {
		unsigned int version;

		version = raw[0] >> 4;
		if (version == 4) {
			struct ip *ip = (struct ip *) raw;

			if (*off + 20 > len
			 || *off + (ip->ip_hl << 2) > len)
				goto truncated;

			*off += (ip->ip_hl << 2);
			return ip;
		} else {
			/* FIXME: Transparently handle v4 and v6 */
			dprint(" (IPv%u not supported)", version);
			return NULL;
		}
	} else if (protocol == IPPROTO_ICMP) {
		if (*off + 8 > len)
			goto truncated;

		*off += 8;
		return raw;
	} else if (protocol == IPPROTO_TCP) {
		/* We don't need the whole TCP header, we just need
		 * source and dest port */
		if (*off + 4 > len)
			goto truncated;

		return raw;
	}

	return NULL;

truncated:
	dprint(" (truncated)");
	return NULL;
}

static void *
ip_get_transhdr(void *pkt, unsigned int len, unsigned int *off,
		int transport_protocol)
{
	struct ip *ip;

	if (!(ip = ip_get_header(IPPROTO_IP, pkt, len, off)))
		return 0;
	if (ip->ip_p != transport_protocol) {
		dprint(" (protocol mismatch)");
		return 0;
	}

	return ip_get_header(transport_protocol, pkt, len, off);
}

/*
 * Helper routine to build IP options
 */
static void
ipopt_init(void)
{
	unsigned int	nr;

	if (af == AF_INET6) {
		if (ngateways)
			fprintf(stderr,
				"Sorry, this traceroute implementation does "
				"not support source routing for IPv6.\n");
		return;
	}

	if (ngateways) {
		ipoptions[0] = IPOPT_NOP; /* pad */
		ipoptions[1] = IPOPT_LSRR;
		ipoptions[2] = 39;
		ipoptions[3] = 4;
		for (nr = 0; nr < ngateways; nr++) {
			struct sockaddr_in	*sin;

			sin = (struct sockaddr_in *) &opt_gateway[nr];
			memcpy(ipoptions + ((nr + 1) << 2),
				&sin->sin_addr, 4);
		}
		ipoptions_len = 40;
	} else
	if (opt_rr) {
		ipoptions[0] = IPOPT_NOP; /* pad */
		ipoptions[1] = IPOPT_RR;
		ipoptions[2] = 39;
		ipoptions[3] = 4;
		ipoptions_len = 40;
	}
}

/*
 * Compute IP checksum
 */
uint32_t
ip_csum_partial(const void *pkt, unsigned int len, uint32_t sum)
{
	const uint16_t *w = (const uint16_t *) pkt;

	while (len > 1) {
		sum += *w++;
		len -= 2;
	}

	if (len)
		sum += *(unsigned char *) w;
	return sum;
}

uint16_t
ip_csum_fold(uint32_t sum)
{
	uint16_t result;

	sum = (sum >> 16) + (sum & 0xffff);
	sum += (sum >> 16);
	result = ~sum;
	if (result == 0)
		result = ~0;
	return result;

}

/*
 * Common socket level functions
 */
static uint16_t
udptcp_bind_port(int fd)
{
	struct sockaddr_storage	addr;
	socklen_t alen;
	int count;

	alen = af_addrlen();
	if (src_addr.ss_family != af) {
		memset(&addr, 0, sizeof(addr));
		addr.ss_family = af;
	} else {
		memcpy(&addr, &src_addr, alen);
	}

	for (count = 0; count < 65536; ++count) {
		uint16_t port = src_port++;

		af_setport((struct sockaddr *) &addr, port);

		if (bind(fd, (struct sockaddr *) &addr, alen) >= 0)
			return port;
		if (errno != EADDRINUSE)
			fatal("unable to bind socket: %m");
	}

	fatal("No free ports available\n");
	return 0;
}

static int
udptcp_connect(int fd, uint16_t port)
{
	struct sockaddr_storage	addr;
	socklen_t alen;

	alen = af_addrlen();
	memcpy(&addr, &dst_addr, alen);
	af_setport((struct sockaddr *) &addr, port);

	if (connect(fd, (struct sockaddr *) &addr, alen) < 0)
		return -1;

	return 0;
}

/*
 * Traceroute over UDP
 * This uses the MSG_ERRQUEUE functionality that lets you
 * do UDP traceroute without needing root privilege.
 */
static int
udp_init_probe(struct hop *hop, struct probe *pb)
{
	int fd, val, len;

	if ((fd = socket(af, SOCK_DGRAM, 0)) < 0)
		fatal("unable to create UDP socket: %m");

	len = sizeof(val);
	if (af == AF_INET) {
		val = 1;
		if (setsockopt(fd, SOL_IP, IP_RECVERR, &val, len) < 0)
			fatal("unable to set SO_RECVERR: %m");
		val = hop->ttl;
		if (setsockopt(fd, SOL_IP, IP_TTL, &val, len) < 0)
			fatal("unable to set TTL: %m");
		val = dst_tos;
		if (setsockopt(fd, SOL_IP, IP_TOS, &val, len) < 0)
			fatal("unable to set TOS: %m");
		val = opt_dontfrag;
		if (setsockopt(fd, SOL_IP, IP_MTU_DISCOVER, &val, len) < 0)
			fatal("unable to set MTU_DISCOVER: %m");
	} else {
		val = 1;
		if (setsockopt(fd, SOL_IPV6, IPV6_RECVERR, &val, len) < 0)
			fatal("unable to set SO_RECVERR: %m");
		val = hop->ttl;
		if (setsockopt(fd, SOL_IPV6, IPV6_UNICAST_HOPS, &val, len) < 0)
			fatal("unable to set IPV6_UNICAST_HOPS: %m");
	}

	val = opt_dontroute;
	if (setsockopt(fd, SOL_SOCKET, SO_DONTROUTE, &val, len) < 0)
		fatal("unable to set SO_DONTROUTE: %m");
	val = 1;
	if (setsockopt(fd, SOL_SOCKET, SO_TIMESTAMP, &val, len) < 0)
		fatal("unable to set SO_TIMESTAMP: %m");

	if (src_device) {
		len = strlen(src_device)+1;
		if (setsockopt(fd, SOL_SOCKET, SO_BINDTODEVICE, src_device, len) < 0)
			fatal("Unable to bind to network interface %s: %m",
				src_device);
	}

	udptcp_bind_port(fd);

	if (udptcp_connect(fd, dst_port++) < 0)
		fatal("Unable to connect to %s: %m", straddr(&dst_addr));

	pb->timeout = now + opt_timeout * 1000;
	pb->hop = hop;
	pb->fd = fd;
	return 0;
}

static inline struct cmsghdr *
cmsg_put(struct cmsghdr *cm, int type, void *data, size_t len)
{
	cm->cmsg_level = SOL_IP;
	cm->cmsg_type  = type;

	memcpy(CMSG_DATA(cm), data, len);
	cm->cmsg_len = CMSG_LEN(len);

	return (struct cmsghdr *) ((caddr_t) cm + cm->cmsg_len);
}

static int
udp_send_probe(struct probe *pb)
{
	struct msghdr	msg;
	unsigned char	control[1024];
	struct cmsghdr	*cm;
	struct iovec	iov;

	memset(&msg, 0, sizeof(msg));
	iov.iov_base	= packet;
	iov.iov_len	= packetsize;

	msg.msg_iov	= &iov;
	msg.msg_iovlen	= 1;

	/* Copy IP options, if specified.  */
	cm = (struct cmsghdr *) control;
	if (ipoptions_len) {
		cm = cmsg_put(cm, IP_RETOPTS, ipoptions, ipoptions_len);
	}

	/* Set source address, if specified */
	if (src_addr.ss_family == AF_INET) {
		struct in_pktinfo info;

		memset(&info, 0, sizeof(info));
		info.ipi_spec_dst = ((struct sockaddr_in *) &src_addr)->sin_addr;

		cm = cmsg_put(cm, IP_PKTINFO, &info, sizeof(info));
	} else
	if (src_addr.ss_family == AF_INET6) {
		struct in6_pktinfo info;

		memset(&info, 0, sizeof(info));
		info.ipi6_addr = ((struct sockaddr_in6 *) &src_addr)->sin6_addr;

		cm = cmsg_put(cm, IPV6_PKTINFO, &info, sizeof(info));
	}

	msg.msg_controllen = ((unsigned char *) cm) - control;
	if (msg.msg_controllen)
		msg.msg_control	= control;

	return sendmsg(pb->fd, &msg, 0);
}

static struct protocol	trace_udp = {
	.ipproto	= IPPROTO_UDP,
	.name		= "UDP",
	.poll_mask	= POLLERR,
	.dst_port	= 33434,
	.init		= NULL,
	.init_probe	= udp_init_probe,
	.send_probe	= udp_send_probe,
	.recv_probe	= common_recv_probe,
};

/*
 * Traceroute over TCP.
 * This is not as elegant as UDP, as there's no errqueue support for
 * TCP in the Linux kernel. So we have to fall back to receiving the
 * ICMP messages in the old-fashioned way.
 */
static void
tcp_init(void)
{
	icmp_init_recvsock();
}

static int
tcp_init_probe(struct hop *hop, struct probe *pb)
{
	pb->timeout = now + opt_timeout * 1000;
	pb->hop = hop;
	pb->fd = -1;
	return 0;
}

static int
tcp_send_probe(struct probe *pb)
{
	struct hop *hop = pb->hop;
	int val, len;

	len = sizeof(val);
	pb->fd = socket(af, SOCK_STREAM, 0);
	if (pb->fd < 0)
		fatal("Unable to create TCP socket: %m");
	if (fcntl(pb->fd, F_SETFL, O_NONBLOCK) < 0)
		fatal("Cannot set socket to nonblocking: %m");

	if (af == AF_INET) {
		val = 1;
		if (setsockopt(pb->fd, SOL_IP, IP_RECVERR, &val, len) < 0)
			fatal("unable to set SO_RECVERR: %m");
		val = hop->ttl;
		if (setsockopt(pb->fd, SOL_IP, IP_TTL, &val, len) < 0)
			fatal("unable to set TTL: %m");
		val = dst_tos;
		if (setsockopt(pb->fd, SOL_IP, IP_TOS, &val, len) < 0)
			fatal("unable to set TOS: %m");
		val = opt_dontfrag;
		if (setsockopt(pb->fd, SOL_IP, IP_MTU_DISCOVER, &val, len) < 0)
			fatal("unable to set MTU_DISCOVER: %m");
	} else {
		val = 1;
		if (setsockopt(pb->fd, SOL_IPV6, IPV6_RECVERR, &val, len) < 0)
			fatal("unable to set SO_RECVERR: %m");
		val = hop->ttl;
		if (setsockopt(pb->fd, SOL_IPV6, IPV6_UNICAST_HOPS, &val, len) < 0)
			fatal("unable to set IPV6_UNICAST_HOPS: %m");
	}

	val = opt_dontroute;
	if (setsockopt(pb->fd, SOL_SOCKET, SO_DONTROUTE, &val, len) < 0)
		fatal("unable to set SO_DONTROUTE: %m");
	val = 1;
	if (setsockopt(pb->fd, SOL_SOCKET, SO_TIMESTAMP, &val, len) < 0)
		fatal("unable to set SO_TIMESTAMP: %m");

	if (src_device) {
		len = strlen(src_device)+1;
		if (setsockopt(pb->fd, SOL_SOCKET, SO_BINDTODEVICE, src_device, len) < 0)
			fatal("Unable to bind to network interface %s: %m",
				src_device);
	}


	pb->seq = udptcp_bind_port(pb->fd);
	if (udptcp_connect(pb->fd, dst_port) == 0) {
		/* we connected... */
		printf("Connected\n");
		return 0;
	} else if (errno != EINPROGRESS) {
		/* We should really just report an error for this hop */
		fatal("Cannot connect: %m");
	} else {
		/* Wait for it to complete */
	}

	return 0;
}

static void
tcp_recv_probe(struct probe *pb)
{
	struct sockaddr addr;
	struct timeval tv;
	int final = 1;

	gettimeofday(&tv, NULL);

	/* Retry the connect() call to reap any error codes */
	dprint("TCP probe %u (ttl %u):", pb->seq, pb->hop->ttl);
	if (udptcp_connect(pb->fd, dst_port) >= 0) {
		/* we made it */
		dprint("connection established\n");
	} else {
		dprint("%m\n");
		switch (errno) {
		case EINPROGRESS:
			/* still pending */
			return;

		case EHOSTUNREACH:
		case ENETUNREACH:
			/* This isn't the final response yet -
			 * we wait for the ICMP. However, we
			 * want to close the socket now. */
			final = 0;
			break;

		default:
			/* Normally we'll see ECONNREFUSED
			 * when the port isn't in use */
			break;
		}
	}

	if (!pb->done)
		pb->recvd_time = tv;

	/* Connecting to an AF_UNSPEC addr on Linux will disconnect
	 * the socket, without going through TIME_WAIT. This is
	 * admittedly a protocol violation, but so is half-open scanning
	 * from an application... */
	addr.sa_family = AF_UNSPEC;
	connect(pb->fd, &addr, sizeof(addr));
	close(pb->fd);
	pb->fd = -1;

	if (final) {
		pb->responder = dst_addr;
		pb->done = 1;

		have_final_response = 1;
		pb->hop->final = 1;
	}
}

struct probe *
tcp_find_probe(void *packet, unsigned int pktlen,
		struct probe **probes, unsigned int nprobes)
{
	struct tcphdr *tcp = packet;
	unsigned int i;
	uint16_t seq;

	dprint(" TCP %u -> %u",
			ntohs(tcp->th_sport),
			ntohs(tcp->th_dport));
	if (ntohs(tcp->th_dport) != dst_port)
		return NULL;

	seq = ntohs(tcp->th_sport);
	for (i = 0; i < nprobes; ++i) {
		if (seq == probes[i]->seq)
			return probes[i];
	}

	return NULL;
}


static struct protocol	trace_tcp = {
	.ipproto	= IPPROTO_TCP,
	.name		= "TCP",
	.poll_mask	= POLLIN|POLLOUT,
	.dst_port	= 80,

	.init		= tcp_init,
	.init_probe	= tcp_init_probe,
	.send_probe	= tcp_send_probe,
	.recv_probe	= tcp_recv_probe,
	.find_probe	= tcp_find_probe,
};

/*
 * Traceroute over ICMP
 */
static int
icmp_init_probe(struct hop *hop, struct probe *pb)
{
	struct icmp	*icmp;
	uint32_t	csum;

	memset(pb->hdr, 0, sizeof(pb->hdr));
	if (af == AF_INET) {
		struct ip *ip = (struct ip *) pb->hdr;

		/* Fill in the IP header. total length
		 * and checksum are done by the kernel */
		ip->ip_v = 4;
		ip->ip_hl = (20 + ipoptions_len) >> 2;
		ip->ip_tos = dst_tos;
		ip->ip_off = opt_dontfrag? htons(IP_DF) : 0;
		ip->ip_ttl = hop->ttl;
		ip->ip_p = IPPROTO_ICMP;
		ip->ip_dst = ((struct sockaddr_in *) &dst_addr)->sin_addr;
		ip->ip_src = ((struct sockaddr_in *) &src_addr)->sin_addr;

		memcpy(ip + 1, ipoptions, ipoptions_len);
		icmp = (struct icmp *) (pb->hdr + ip->ip_hl);
		pb->hdrlen = 20 + ipoptions_len;
	} else {
		fatal("ICMP probes over IPv6 not yet implemented\n");
	}

	pb->seq = icmp_sequence++;
	pb->hdrlen += 8;
	icmp->icmp_type = ICMP_ECHO;
	icmp->icmp_id = icmp_ident;
	icmp->icmp_seq = htons(pb->seq);

	icmp->icmp_cksum = 0;
	csum = ip_csum_partial(icmp, 8, 0);
	csum = ip_csum_partial(packet, packetsize, csum);
	icmp->icmp_cksum = ip_csum_fold(csum);

	csum = ip_csum_partial(icmp, 8, 0);
	csum = ip_csum_partial(packet, packetsize, csum);
	csum = ip_csum_fold(csum);

	pb->timeout = now + opt_timeout * 1000;
	pb->hop = hop;
	pb->fd = -1;
	return 0;
}

static int
icmp_send_probe(struct probe *pb)
{
	struct msghdr	msg;
	struct iovec	iov[2];

	memset(&msg, 0, sizeof(msg));
	iov[0].iov_base	= pb->hdr;
	iov[0].iov_len	= pb->hdrlen;
	iov[1].iov_base	= packet;
	iov[1].iov_len	= packetsize;

	msg.msg_iov	= iov;
	msg.msg_iovlen	= 2;
	msg.msg_name	= &dst_addr;
	msg.msg_namelen	= af_len;

	return sendmsg(icmp_sendsock, &msg, 0);
}

struct probe *
icmp_find_probe(void *packet, unsigned int pktlen,
		struct probe **probes, unsigned int nprobes)
{
	struct icmp *icmp = packet;
	unsigned int i;
	uint16_t seq;

	if (icmp->icmp_type != ICMP_ECHO
	 && icmp->icmp_type != ICMP_ECHOREPLY)
		return NULL;

	if (icmp->icmp_id != icmp_ident)
		return NULL;

	seq = ntohs(icmp->icmp_seq);
	for (i = 0; i < nprobes; ++i) {
		if (seq == probes[i]->seq)
			return probes[i];
	}

	return NULL;
}

static int
icmp_recv(struct probe **probes, int nprobes)
{
	struct response	response;
	uint32_t	packet[64];
	void		*transhdr;
	int		len;
	unsigned int	offset = 0;
	struct icmp	*icmp = NULL;
	struct probe	*pb;
	const char	*queue = "";

	memset(&response, 0, sizeof(response));
	response.buf = packet;
	response.bufsz = sizeof(packet);

	/* First, try to receive any pending messages from the
	 * socket's error queue.  If there are non, try receiving
	 * from the regular message queue. */
	len = probe_recverr(icmp_recvsock, &response, MSG_ERRQUEUE);
	if (len >= 0) {
		queue = " (MSG_ERRQUEUE)";
	} else {
		len = probe_recverr(icmp_recvsock, &response, 0);
	}

	if (len < 0)
		return -1;

	dprint("Received %u bytes%s", len, queue);
	if (response.ee.ee_origin == SO_EE_ORIGIN_ICMP) {
		dprint(" from error queue");

		/* This is the original packet causing the error */
		transhdr = icmp = (struct icmp *) packet;
	} else {
		icmp = ip_get_transhdr(packet, len, &offset, IPPROTO_ICMP);
		if (icmp == NULL)
			goto out;

		if (icmp->icmp_type == ICMP_DEST_UNREACH
		 || icmp->icmp_type == ICMP_TIME_EXCEEDED
		 || icmp->icmp_type == ICMP_PARAMETERPROB) {
			/* ICMP error - dissect the enclosed packet */
			transhdr = ip_get_transhdr(packet, len, &offset, proto->ipproto);
			if (!transhdr)
				goto out;

			response.ee.ee_origin = SO_EE_ORIGIN_ICMP;
			response.ee.ee_type = icmp->icmp_type;
			response.ee.ee_code = icmp->icmp_code;
		} else {
			/* ICMP is the transport */
			transhdr = icmp;
		}
	}

	if (opt_debug) {
		switch (icmp->icmp_type) {
		case ICMP_ECHO:
			dprint(" echo 0x%x", ntohs(icmp->icmp_seq));
			break;
		case ICMP_ECHOREPLY:
			dprint(" echo reply 0x%x", ntohs(icmp->icmp_seq));
			break;
		case ICMP_TIME_EXCEEDED:
			dprint(" time_exceeded");
			break;
		default:
			dprint(" message type %u", icmp->icmp_type);
		}
	}

	assert(len >= offset);

	pb = proto->find_probe(transhdr, 8, probes, nprobes);
	if (pb == NULL) {
		dprint(": noise");
	} else {
		dprint(": response to probe %u", pb->seq);
		pb->responder = response.sender;
		pb->recvd_time = response.ts;
		probe_seterror(pb, &response.ee);
		pb->done = 1;

		probe_reduce_timeout(pb->hop->ttl);
	}

out:
	dprint("\n");
	return 0;
}

static void
icmp_init_recvsock(void)
{
	int	fd, val;

	/* Create the receive socket */
	if ((fd = socket(af, SOCK_RAW, IPPROTO_ICMP)) < 0)
		fatal("unable to create ICMP recv socket: %m");
	icmp_recvsock = fd;
	fcntl(fd, F_SETFL, O_NONBLOCK);

	val = 1;
	if (setsockopt(fd, SOL_SOCKET, SO_TIMESTAMP, &val, sizeof(val)) < 0)
		fatal("unable to set SO_TIMESTAMP: %m");
	if (af == AF_INET) {
		if (setsockopt(fd, SOL_IP, IP_RECVERR, &val, sizeof(val)) < 0)
			fatal("unable to set SO_RECVERR: %m");
	} else {
		if (setsockopt(fd, SOL_IPV6, IPV6_RECVERR, &val, sizeof(val)) < 0)
			fatal("unable to set SO_RECVERR: %m");
	}

	if (src_device) {
		unsigned int len = strlen(src_device)+1;
		if (setsockopt(fd, SOL_SOCKET, SO_BINDTODEVICE, src_device, len) < 0)
			fatal("Unable to bind to network interface %s: %m",
				src_device);
	}
}

static void
icmp_init(void)
{
	int	fd, val;

	/* Create the send socket */
	if ((fd = socket(af, SOCK_RAW, IPPROTO_RAW)) < 0)
		fatal("unable to create ICMP send socket: %m");
	icmp_sendsock = fd;

#if 0
	if (src_addr.ss_family != AF_UNSPEC
	 && bind(fd, (struct sockaddr *) &src_addr, af_len) < 0)
		fatal("unable to bind ICMP send socket: %m");
	if (connect(fd, (struct sockaddr *) &dst_addr, af_len) < 0)
		fatal("unable to connect ICMP send socket: %m");
#endif

	val = opt_dontroute;
	if (setsockopt(fd, SOL_SOCKET, SO_DONTROUTE, &val, sizeof(val)) < 0)
		fatal("unable to set SO_DONTROUTE: %m");
	val = 1;
	if (af == AF_INET) {
		if (setsockopt(fd, SOL_IP, IP_HDRINCL, &val, sizeof(val)) < 0)
			fatal("unable to set SO_TIMESTAMP: %m");
	} else {
		/* Not yet */
	}

	if (src_device) {
		unsigned int len = strlen(src_device)+1;
		if (setsockopt(fd, SOL_SOCKET, SO_BINDTODEVICE, src_device, len) < 0)
			fatal("Unable to bind to network interface %s: %m",
				src_device);
	}

	/* Create the receive socket */
	icmp_init_recvsock();

	icmp_ident = (getpid() & 0xffff) | 0x8000;
}

static struct protocol	trace_icmp = {
	.ipproto	= IPPROTO_ICMP,
	.name		= "ICMP",
	.poll_mask	= 0,
	.dst_port	= 33434,

	.init		= icmp_init,
	.init_probe	= icmp_init_probe,
	.send_probe	= icmp_send_probe,
	.recv_probe	= common_recv_probe,
	.find_probe	= icmp_find_probe,
};

